<!--
Use md2man (https://github.com/sunaku/md2man) to generate the man file like this.
$ md2man-roff man.md > teip.1
-->
TEIP 1 "APR 2022" "User Commands" ""
=======================================

NAME
----

teip - Highly efficient "Masking tape" for standard input

SYNOPSIS
--------

`teip` -g <*pattern*> [-oGsvz] [--] [<*command*>...]

`teip` -f <*list*> [-d <*delimiter*> | -D <*pattern*>] [-svz] [--] [<*command*>...]

`teip` -c <*list*> [-svz] [--] [<*command*>...]

`teip` --help | --version

DESCRIPTION
-----------
Only a selected part of standard input is passed to any command for execution.

OPTIONS
-------
`--help`
  Display this help and exit

`--version`
  Show version and exit

`-g` <*pattern*>
  Select lines that match the regular expression <*pattern*>

`-o`
  -g selects only matched parts

`-G`
  -g adopts Oniguruma regular expressions

`-f` <*list*>
  Select only these white-space separated fields

`-d` <*delimiter*>
  Use <*delimiter*> for field delimiter of -f

`-D` <*pattern*>
  Use a regular expression <*pattern*> for field delimiter of -f

`-c` <*list*>
  Select only these characters

`-e` <*string*>
  Execute <*string*> on another process that will receive identical standard input as the teip, and numbers given by the result are used as line numbers for bypassing

`-s`
  Execute command for each selected part

`-v`
  Invert the sense of selecting

`-z`
  NUL is used as line delimiter instead of the newline

### *command*

*command* is the command and its arguments that `teip` executes.
*command* must print a single line of result for each line of the input.
In the simplest example, the cat(1) command always succeeds.
Because the cat prints the same number of lines against the input.

```
$ echo ABCDEF | teip -og . -- cat
ABCDEF
```

sed(1) works with the typical pattern.

```
$ echo ABCDEF | teip -og . -- sed 's/[ADF]/@/'
@BC@E@
```

If the rule is not satisfied, the result will be inconsistent.
For example, the grep(1) may fail. Here is an example.

```
$ echo ABCDEF | teip -og . -- grep '[ABC]'
ABC
teip: Output of given command is exhausted
```

`teip` could not get the result corresponding to D, E, and F. That is why the example fails.
If such the inconsistency occurs, `teip` will exit with the error message. Then, the exit status will be 1.

```
$ echo $?
1
```

If *command* is not given, `teip` shows how standard input will be devided into chunks.

```
$ echo ABCDEF | teip -og .
[A][B][C][D][E][F]
```

### *list*

*list* is an expression to specify the range of fields or characters.
The notation is compatible with the one used in cut(1). Refer to the cut(1) manual in detail.
Here are some examples.

Select 1st, 3rd, and 5th fields.

```
$ echo 1 2 3 4 5 | teip -f 1,3,5 -- sed 's/./@/'
@ 2 @ 4 @
```

Select fields between 2nd and 4th.

```
$ echo 1 2 3 4 5 | teip -f 2-4 -- sed 's/./@/'
1 @ @ @ 5
```

Select all the fields after 3rd.

```
$ echo 1 2 3 4 5 | teip -f 3- -- sed 's/./@/'
1 2 @ @ @
```

Select all the fields before 4th.

```
$ echo 1 2 3 4 5 | teip -f -4 -- sed 's/./@/'
@ @ @ @ 5
```

### *pattern*

*pattern* is a regular expression whose grammar follows "regex crate".
Refer to the link in *SEE ALSO* about the details.

### Necessity of **--**

`teip` interprets arguments after `--` as *command* and its argument.

If **--** is omitted, the command fails in this example.

```
$ echo "100 200 300 400" | teip -f 3 cut -c 1
teip: Invalid arguments.
```

This is because the `cut` uses the `-c` option. The option of the same name is also provided by `teip`, which is confusing.

```
$ echo "100 200 300 400" | teip -f 3 -- cut -c 1
100 200 3 400
```

### External execution for match offloading (`-e`)

With `-e`, you can use the external commands you are familiar with to specify the range of holes.
`-e` allows you to specify the shell pipeline as a string. This pipeline is executed in `/bin/sh`.

For example, with a pipeline `echo 3` that outputs `3`, then only the third line will be bypassed.

```bash
$ echo -e 'AAA\nBBB\nCCC' | teip -e 'echo 3'
AAA
BBB
[CCC]
```

It works even if the output is somewhat 'dirty'.
For example, if any spaces or tab characters are included at the beginning of a line, they are ignored.
Also, once a number is given, it does not matter if there are non-numerical characters to the right of the number.

```bash
$ echo -e 'AAA\nBBB\nCCC' | teip -e 'echo " 3"'
AAA
BBB
[CCC]
$ echo -e 'AAA\nBBB\nCCC' | teip -e 'echo " 3:testtest"'
AAA
BBB
[CCC]
```

Technically, the first captured group in the regular expression `^\s*([0-9]+)` is interpreted as a line number.
`-e` will also recognize multiple numbers if the pipeline provides multiple lines of numbers.

```
$ echo -e 'AAA\nBBB\nCCC\nDDD\nEEE\nFFF' | teip -e 'seq 1 2 10' -- sed 's/. /@/g'
@@@
BBB
@@@
DDD
@@@
FFF
```

Note that the order of the numbers must be in ascending order.

The pipeline obtains identical standard input as `teip`.
The following command is a `grep` command that prints **the line numbers of the line containing the string "CCC" and the two lines after it**.

```
$ echo -e 'AAA\nBBB\nCCC\nDDD\nEEE\nFFF' | grep -n -A 2 CCC
3:CCC
4-DDD
5-EEE
```

If you give this command to `-e`, you can punch holes in **the line containing the string "CCC" and the two lines after it**.

```
$ echo -e 'AAA\nBBB\nCCC\nDDD\nEEE\nFFF' | teip -e 'grep -n -A 2 CCC'
AAA
BBB
[CCC]
[DDD]
[EEE]
FFF
```

GNU `sed` has `=`, which prints the line number being processed.
Below is an example of how to drill from the line containing "BBB" to the line containing "EEE".

```
$ echo -e 'AAA\nBBB\nCCC\nDDD\nEEE\nFFF' | teip -e 'sed -n "/BBB/,/EEE/="'
AAA
[BBB]
[CCC]
[DDD]
[EEE]
FFF
```

Of course, similar operations can also be done with `awk`.

```
$ echo -e 'AAA\nBBB\nCCC\nDDD\nEEE\nFFF' | teip -e 'awk "/BBB/,/EEE/{print NR}"'
```

The following is an example of combining the commands `nl` and `tail`.
You can only make holes in the last three lines of input.

```
$ echo -e 'AAA\nBBB\nCCC\nDDD\nEEE\nFFF' | teip -e 'nl -ba | tail -n 3'
AAA
BBB
CCC
[DDD]
[EEE]
[FFF]
```

The `-e` argument is a single string.
Therefore, pipe `|` and other symbols can be used as it is.


EXAMPLES
-------

Edit 2nd, 3rd, and 4th columns in the CSV file

```
$ cat file.csv | teip -f 2-4 -d , -- sed 's/./@/g'
```

Convert timestamps in /var/log/secure to UNIX time

```
$ cat /var/log/secure | teip -c 1-15 -- date -f- +%s
```

Edit the line containing 'hello' and the three lines before and after it

```
$ cat access.log | teip -e 'grep -n -C 3 hello' -- sed 's/./@/g'
```

SEE ALSO
--------

### Manual pages
cut(1)

### Full documentation
<https://github.com/greymd/teip>

### Regular expression
https://docs.rs/regex/

### Regular expression (Oniguruma)
https://github.com/kkos/oniguruma/blob/master/doc/RE

AUTHOR AND COPYRIGHT
------

Copyright (c) 2022 Yamada, Yasuhiro <yamada@gr3.ie> Released under the MIT License.
